//==============================================================================================================================
// Tonemapping Operator (human) 10.0.hlsl
// CricWood "Bieder"  Christopher Maiwald, 20082010
//		E-Mail	CricWood@NerdShack.com
//		ICQ		342 173 470
// Codepage: Windows-1252; text width: 128 chars; tab size: 4 spaces
// File comments are in English language.
//==============================================================================================================================



// Direct3D 10 / 11 difference: Direct3D 10.0 can not handle multi-sampled textures in the same, easy way Direct3D 11 can.
//	Therefor, multi-sampling is not available in the Direct3D 10 version. However, to avoid two seperate paths in the executable
//	(creating a non-multisampled 2D-texture and a TEX2D SRV vs. creating a multisampled 2D-texture and a TEX2DMS SRV), this D3D
//	10.0 shader declares a probably multi-sampled texture but uses only one of its samples.
Texture2DMS<float3, 1> g_Input	:register(t0);



cbuffer g_Constants				:register(b0) {
	float3	g_Exposure				:packoffset(c0);
	// Random numbers as a seed for the dithering's random number generator. Should be between 4 (very poor results are
	//	generated below) and 100k (to avoid integer overruns).
	uint4	g_RandomSeed			: packoffset(c1);
};

static const float VisualMinimum	= 0.000003f;
static const float EigengrauMaximum	= 0.001f;
static const float MesopicMinimum	= 0.002f;
static const float PhotopicMinimum	= 2.00f;


//------------------------------------------------------------------------------------------------------------------------------
// Returns pseudo-random noise in the range (0.0, 1.0] for the given position.
//------------------------------------------------------------------------------------------------------------------------------
const float NoiseFor(
	in const uint2 PixelsPosition
) {
	// A pair of values is created which differs per frame, and per texel:

	// Differs per frame: the random seed.
	// Differs per texel: the linearized texel index.
	const uint PixelsIndexBase	= 661 * PixelsPosition.y + PixelsPosition.x;
	const uint PixelsIndex		= g_RandomSeed.r + PixelsIndexBase * 41;

	// Differs per frame: the random seed.
	// Differs per texel: the number of texels in the rect from (0|0) to "TexelsPosition".
	const uint PixelsAreaBase	= PixelsPosition.x * PixelsPosition.y;
	const uint PixelsArea		= g_RandomSeed.r + PixelsAreaBase * 23;

	// XOR both values for pseudo-random noise. Then clamp it to (0.0, 1.0].
	return float((PixelsIndex ^ PixelsArea) % 661) / 661.0f;
}

//------------------------------------------------------------------------------------------------------------------------------
// Returns pseudo-random Eigengrau with a frequency of 2 pixels and a maximal luminance of 0.04 cd/m for the given pixel
//	position. See http://en.wikipedia.org/wiki/Eigengrau.
//------------------------------------------------------------------------------------------------------------------------------
const float EigengrauFor(
	in const uint2 PixelsPosition
) {
	// Produce noise with a frequency of 22 pixels through bilinear interpolation.
	const float UpperLeft  = NoiseFor(PixelsPosition + float2(0, 0));
	const float UpperRight = NoiseFor(PixelsPosition + float2(1, 0));
	const float LowerLeft  = NoiseFor(PixelsPosition + float2(0, 1));
	const float LowerRight = NoiseFor(PixelsPosition + float2(1, 1));
	const float Noise = 0.25f * ((UpperLeft + UpperRight) + (LowerLeft + LowerRight));

	// f noise -> f noise
	return EigengrauMaximum * pow(Noise, 2.0f);
}

//------------------------------------------------------------------------------------------------------------------------------
// The pixel shader.
//------------------------------------------------------------------------------------------------------------------------------
const float3 ProcessPixel(
	in const float4 ItsPosition : SV_Position
) :SV_Target0 {
	const float3 Sample = max(VisualMinimum, g_Input.Load(int2(ItsPosition.xy), 0) + EigengrauFor(ItsPosition.xy));

	const float3 MesopicToPhotopicRatio = pow(saturate((Sample - MesopicMinimum) / (PhotopicMinimum - MesopicMinimum)), 0.7f);
	const float3 ScotopicToMesopicRatio = saturate(Sample / MesopicMinimum);

	const float3 PhotopicColor = saturate(1.0f - exp(-g_Exposure * Sample));
	const float3 MesopicColor  = dot(float3(0.114f, 0.587f, 0.299f), PhotopicColor) * float3(0.75f, 0.95f, 1.30f);
	const float3 ScotopicColor = dot(float3(0.114f, 0.299f, 0.587f), PhotopicColor) * float3(0.50f, 0.65f, 1.85f);
//	const float3 PhotopicColor = float3(1.f, 0.f, 0.f);
//	const float3 MesopicColor  = float3(0.f, 1.f, 0.f);
//	const float3 ScotopicColor = float3(0.f, 0.f, 1.f);

	return lerp(lerp(ScotopicColor, MesopicColor, ScotopicToMesopicRatio), PhotopicColor, MesopicToPhotopicRatio);
}
